body {
  font-family: "Helvetica Neue", Helvetica, Arial, sans-serif;
  font-size: 13px;
  line-height: 1.42857143;
  color: #000;
  background-color: #fff;

  margin: 0 0 50px;
  /* bottom = footer height */
  padding: 25px;
}

html {
  position: relative;
  min-height: 100%;
}

footer {
  position: absolute;
  left: 0;
  bottom: 0;
  width: 100%;
  padding-top: 1em;
  padding-bottom: 1em;
  padding-left: 25px;
  padding-right: 25px;
  background-color: #f5f5f5;
  overflow: hidden;
}

textarea.sos-source {
  line-height: 1.21429em;
  font-size: 14px;
  background: none;
  width: 100%;
  border: none;
  font-family: monospace;
  height: auto;
  padding: 0.4em;
  outline: none;
  position: relative;
  resize: none;
  overflow: hidden;
  vertical-align: text-top;
}

.code {
  font-family: monotone;
}
h1, h2, h3, h4, h5 {
 font-family: "Source Sans Pro", sans-serif;
}
.h1:first-child,
.h2:first-child,
.h3:first-child,
.h4:first-child,
.h5:first-child,
.h6:first-child {
 margin-top: 0.2em;
}

.h1 {
 color: #126dce;
 font-size: 220%;
 text-align: center;
 font-weight: lighter;
}
.h2 {
 text-align: left;
 font-size: 170%;
 color: #126dce;
 font-style: normal;
 font-weight: lighter;
}
.h3 {
 font-size: 150%;
 color: #126dce;
 font-weight: lighter;
 text-decoration: italic;
 font-style: normal;
}
.h4 {
 font-size: 120%;
 color: #126dce;
 font-weight: underline;
 font-style: normal;
}


table {
  /* margin-left: auto; */
  margin-right: auto;
  border: none;
  border-collapse: collapse;
  border-spacing: 0;
  color: black;
  font-size: 12px;
  table-layout: fixed;
}
tr,
th,
td {
  text-align: right;
  vertical-align: middle;
  padding: 0.5em 0.5em;
  line-height: normal;
  white-space: normal;
  max-width: none;
  border: none;
}
td,
th {
  text-align: left;
  vertical-align: middle;
  padding: 4px;
  display: table-cell;
}
th {
  font-weight: bold;
}
* + table {
  margin-top: 1em;
}

tbody tr:nth-child(odd) {
  background: #f5f5f5;
}
tbody tr:hover {
  background: rgba(66, 165, 245, 0.2);
}

img.dag-image {
  max-width: 80%;
}

.timeline-col {
  width : 300px;
  text-align: center;
}

.timeline-cell {
  display: flex;
  width: 100%;
  height: 1em;
  margin: 0px;
	overflow: hidden;
}
.timeline-before {
  float: left;
	background-color: #FEF9E7;
  height: 1em;
  margin: 0px;
}
.timeline-during {
  float: left;
	background-color: green;
	min-width: 1px;
  height: 1em;
  margin: 0px;
}
.timeline-after {
  float: left;
	background-color: #FEF9E7;
  height: 1em;
  margin: 0px;
}

.master_workflow th {
  font-weight: bold;
}

.subworkflow th {
  font-style: normal;
}

.dropdown-menu li
{
  padding-left: 5px;
  padding-right: 100px;
  white-space: nowrap;
  font-size: small;
  font-family: monospace;
}

.dropdown-menu > li > .filesize {
  position: absolute;
  right: 5px;
}

a {
  text-decoration: none;
}

a:link, a:visited {
  color: black;
}

a:hover {
  color: red;
}

tr.subworkflow.step-row {
  display: none;
}

.filelist {
  cursor: pointer;
}

.code {
  color: black!important;
}

#sourcecode {
  display:none;
}

.sos-logo img  {
  height: 2em;
}

.fa-external-link-alt {
  color: gray;
  font-size: 0.5em;
}

.cm-sos-interpolated {
  background-color: #EDD5F3;
}
.cm-sos-sigil {
  background-color: #EDD5F3;
}

.file {
  position: relative;
  margin-top: 16px;
  margin-bottom: 16px;
  border: 1px solid #ddd;
  border-radius: 3px;
}

.file .fileheader {
  padding: 5px 10px;
  background-color: #fafbfc;
  border-bottom: 1px solid #e1e4e8;
  border-top-left-radius: 2px;
  border-top-right-radius: 2px;
}

.file .fileinfo {
  font-family: "SFMono-Regular",Consolas,"Liberation Mono",Menlo,Courier,monospace;
  font-size: 12px;
  line-height: 32px;
}

.file .file-info-divider {
  display: inline-block;
  width: 1px;
  height: 18px;
  margin-right: 3px;
  margin-left: 3px;
  vertical-align: middle;
  background-color: #ddd;
}

.file .filecontent {
  overflow-x: auto;
  overflow-y: hidden;
  border-bottom-right-radius: 3px;
  border-bottom-left-radius: 3px;
}

.file .CodeMirror {
  border: 1px solid #eee;
  height: auto;
}

pre.section-header.CodeMirror-line {
  background-color: #f7f7f7;
  border-top: 1px solid #ddd;
}
